# frozen_string_literal: true

# Legende
# Les script a plusieurs parties indique que les bash (Batch dans l'app)
# peuvent s'initier avec differents types de source
# Le besoin est donc d'isoler la logique associee au modele actuel
# ne pas oublier d'indexer les champs simples

module Hermes
  # Status class
  class Status
    include Hermes::Concerns::CommonFieldsConcern
    field :reference_model_name, type: String
    field :status_label, type: String
    field :motive, type: String
    field :title, type: String
    field :object_id, type: String
    field :performer_id, type: BSON::ObjectId
    field :closed_by, type: BSON::ObjectId
    field :closed_at, type: DateTime
    field :closing, type: BSON::ObjectId

    belongs_to :applicant, class_name: 'Hermes::Applicant'
    belongs_to :housing_request, class_name: 'Hermes::HousingRequest'
    belongs_to :housing_offer, class_name: 'Hermes::HousingOffer'
    belongs_to :housing_application, class_name: 'Hermes::HousingApplication'
    belongs_to :entity, class_name: 'Hermes::Entity'
    belongs_to :lessor, class_name: 'Hermes::Lessor'

    before_create :before_create_actions
    after_create :after_create_actions

    writer_roles 'Superadmin', 'Administrator'
    reader_roles 'Manager'
    owner_field :user_id

    CLASSES = [
      Hermes::Applicant,
      Hermes::HousingRequest,
      Hermes::HousingOffer,
      Hermes::HousingApplication,
      Hermes::Entity,
      Hermes::Lessor
    ]

    def before_create_actions
      handle_reference_model_name
      check_if_closing
    end

    def after_create_actions
      update_object_status
      update_closing if closing.present?
    end

    def handle_reference_model_name
      CLASSES.any? do |model|
        if object = model.find_by(id: object_id)
          class_name = object.class.name
          send("#{class_name.demodulize.underscore}_id=", object.id)
          self.reference_model_name = class_name
        end
      end
    end

    def check_if_closing
      if (
        closing = Hermes::Status.only(:object_id, :status_label, :created_at)
                                .where(object_id: object_id, status_label: status_label)
                                .order(created_at: :desc)
                                .limit(1)
                                .first
      )
        self.closing = closing.id
      end
    end

    def update_object_status
      if (reference_model_name && instance = reference_model_name.to_s.constantize.find_by(id: object_id))
        instance[status_label] = title
        instance["#{status_label}_id"] = id
        instance["#{status_label}_updated_at"] = created_at
        instance["old_#{status_label}"] = Hermes::Status.find(closing).title if closing
        instance.save!
      end
    end

    def update_closing
      c = Hermes::Status.find(closing)
      c.closed_at = DateTime.current
      c.closed_by = id
      c.save!
    end
  end
end
