require 'rails_helper'

RSpec.describe Hermes::HousingOffer, type: :model do
  it 'has a valid factory' do
    expect(FactoryBot.build(:housing_offer)).to be_valid
  end

  describe 'validate save side entities relationships effect' do
    let!(:entity1) { create(:entity, siret: 'test1') }
    let!(:entity2) { create(:entity, siret: 'test2') }
    let!(:entity3) { create(:entity, siret: 'test3') }

    let!(:offer) do
      create(
        :housing_offer,
        entities_siret: ['test1', 'test3']
      )
    end

    it 'should has valid relationships' do
      expect(Hermes::HousingEntityOffer.where(housing_offer: offer).count).to be 2
      expect(offer.entity_ids).to include(entity1.id).and(include(entity3.id))
      expect(offer.readers).to include("g_#{entity1.supervisor_group.id}").and(include("g_#{entity3.supervisor_group.id}"))
      expect(offer.writers).to include("g_#{entity1.manager_group.id}").and(include("g_#{entity3.manager_group.id}"))
    end

    describe 'changing entities_siret' do
      before :each do
        offer.entities_siret = ['test1', 'test2']
        offer.save
      end

      it 'should has valid relationships' do
        expect(Hermes::HousingEntityOffer.where(housing_offer: offer).count).to be 2
        expect(Hermes::HousingEntityOffer.unscoped.where(housing_offer: offer).count).to be 3
        expect(offer.reload.entity_ids).to include(entity1.id).and(include(entity2.id))
        expect(offer.reload.readers).not_to include("g_#{entity3.supervisor_group.id}")
        expect(offer.reload.writers).not_to include("g_#{entity3.manager_group.id}")
        expect(offer.reload.readers).to include("g_#{entity1.supervisor_group.id}").and(include("g_#{entity2.supervisor_group.id}"))
        expect(offer.reload.writers).to include("g_#{entity1.manager_group.id}").and(include("g_#{entity2.manager_group.id}"))
      end
    end
  end

  describe 'validate save side charges effect' do
    let!(:basic_housing_offer) do
      create(
        :housing_offer,
        rental_charges: 24.5,
        parking_charges: 12.3,
        rent_amount: 78.0
      )
    end

    it 'put referent' do
      expect(basic_housing_offer.total_charges).to(
        eq(basic_housing_offer.attributes.slice(
          *%w[
            parking_charges
            rental_charges
          ]
        ).values.sum)
      )
      expect(basic_housing_offer.rent_with_charges).to(
        eq(basic_housing_offer.rent_amount + basic_housing_offer.total_charges)
      )
    end
  end

  describe 'validate save is_mad side effect' do
    let!(:housing_offer1) do
      create(
        :housing_offer,
        lessor_status: 'PLACE',
        offer_status: "EN ATTENTE COMMERCIALISATION"
      )
    end

    let!(:housing_offer2) do
      create(
        :housing_offer,
        lessor_status: 'RETIRE'
      )
    end

    let!(:housing_offer3) do
      create(
        :housing_offer,
        offer_status: 'RETIRE PAR LE BAILLEUR'
      )
    end

    let!(:housing_offer4) do
      create(
        :housing_offer,
        offer_status: 'LOGEMENT PLACE'
      )
    end

    it 'put referent' do
      expect((1..4).map { |i| send("housing_offer#{i}").is_mad }).to all(be false)
    end

    describe 'when status gets changed to mad situation' do
      before :each do
        housing_offer1.update(lessor_status: 'EN ATTENTE PUBLICATION')
      end

      it 'should has the right status' do
        expect(housing_offer1.reload.lessor_status).to eq 'EN ATTENTE PUBLICATION'
      end

      it 'should has updated correctly' do
        offer1 = housing_offer1.reload
        expect(offer1.is_mad).to be true
        expect(offer1.is_mad_updated_at).to_not eq offer1.created_at
      end
    end
  end

  context 'with adequate data for score' do
    #let!(:admin1) { create(:administrator) }

    let!(:threshold1) do
      create(
        :threshold,
        type: 'PLA LM',
        thresholds_map: {
          'Paris et communes limitrophes': {
            'thresholds_list': [
              13_050,
              21_272,
              27_883,
              30_521,
              36_307,
              40_859
            ],
            'floors_list': [
              3_050,
              11_272,
              17_883,
              20_521,
              26_307,
              30_859
            ],
            'another_person': 4552
          },
          'IDF hors Paris et Communes limitrophes': {
            'thresholds_list': [
              13_050,
              21_272,
              25_569,
              28_075,
              33_238,
              37_401
            ],
            'floors_list': [
              3_050,
              11_272,
              17_883,
              20_521,
              26_307,
              30_859
            ],
            'another_person': 4166
          },
          'Autres régions': {
            'thresholds_list': [
              11_342,
              16_525,
              19_872,
              22_111,
              25_870,
              29_155
            ],
            'floors_list': [
              3_050,
              11_272,
              17_883,
              20_521,
              26_307,
              30_859
            ],
            'another_person': 3252
          }
        }
      )
    end

    let(:entity1) { create(:entity) }

    let!(:user1) { create(:user) }
    let!(:applicant1) do
      create(
        :applicant,
        sne_profile_creation_date: '2016-09-24',
        main_user_id: user1.id,
        amount_year_minus_two: 9787.24,
        entity: entity1,
        monthly_incomes: {
          'Salaire': 1045.23,
          'APL': 321.62
        }
      )
    end
    let!(:request1) do
      create(
        :housing_request,
        is_active: true,
        main_applicant: applicant1,
        seeked_districts: ['34507'],
        seeked_departments: ['34'],
        entity_validated: true,
        priority_points_indexes: [30]
      )
    end
    let!(:nud1) do
      create(
        :nud,
        housing_request: request1,
        department: '34',
        creation_date: '2016-09-24'
      )
    end

    let!(:user2) { create(:user) }
    let!(:applicant2) do
      create(
        :applicant,
        sne_profile_creation_date: '2017-09-24',
        main_user_id: user1.id,
        amount_year_minus_two: 9787.24,
        entity: entity1,
        monthly_incomes: {
          'Salaire': 1045.23,
          'APL': 321.62
        }
      )
    end
    let!(:request2) do
      create(
        :housing_request,
        is_active: true,
        main_applicant: applicant2,
        seeked_districts: ['34507'],
        seeked_departments: ['34'],
        entity_validated: true,
        priority_points_indexes: [40]
      )
    end
    let!(:nud2) do
      create(
        :nud,
        housing_request: request2,
        department: '34',
        creation_date: '2017-09-24'
      )
    end

    let!(:user3) { create(:user) }
    let!(:applicant3) do
      create(
        :applicant,
        sne_profile_creation_date: '2018-09-24',
        main_user_id: user3.id,
        amount_year_minus_two: 9787.24,
        entity: entity1,
        monthly_incomes: {
          'Salaire': 1045.23,
          'APL': 321.62
        }
      )
    end
    let!(:request3) do
      create(
        :housing_request,
        is_active: true,
        main_applicant: applicant3,
        seeked_districts: ['34507'],
        seeked_departments: ['34'],
        entity_validated: true,
        priority_points_indexes: [30]
      )
    end
    let!(:nud3) do
      create(
        :nud,
        housing_request: request3,
        department: '34',
        creation_date: '2018-09-24'
      )
    end

    let!(:user4) { create(:user) }
    let!(:applicant4) do
      create(
        :applicant,
        sne_profile_creation_date: '2015-09-24',
        main_user_id: user4.id,
        amount_year_minus_two: 9787.24,
        entity: entity1,
        monthly_incomes: {
          'Salaire': 1045.23,
          'APL': 321.62
        }
      )
    end
    let!(:request4) do
      create(
        :housing_request,
        is_active: true,
        main_applicant: applicant4,
        seeked_districts: ['34507'],
        seeked_departments: ['34'],
        entity_validated: true,
        priority_points_indexes: [40]
      )
    end
    let!(:nud4) do
      create(
        :nud,
        housing_request: request4,
        department: '34',
        creation_date: '2015-09-24'
      )
    end

    let!(:offer1) do
      create(
        :housing_offer,
        typology: 'T2',
        rent_amount: 234.32,
        rental_charges: 42.15,
        parking_charges: 12.71,
        surface: 29,
        offer_status: 'LOGEMENT PUBLIE',
        threshold_type: 'PLA LM',
        department: '34',
        threshold_subtype: 'IDF hors Paris et Communes limitrophes'
      )
    end

    let!(:offer2) do
      create(
        :housing_offer,
        typology: 'T2',
        rent_amount: 434.32,
        rental_charges: 53.12,
        parking_charges: 22.43,
        surface: 34,
        offer_status: 'LOGEMENT PUBLIE',
        threshold_type: 'PLA LM',
        department: '34',
        threshold_subtype: 'Autres régions'
      )
    end

    describe 'handles queued jobs' do
      it 'should have jobs and right quantity of scores' do
        expect(enqueued_jobs).to be_present
        expect(enqueued_jobs.size).to be > 0
        perform_enqueued_jobs do
          enqueued_jobs.each do |job|
            job[:job].perform_now(*job[:args])
          end
        end

        scores = Hermes::Score.where(housing_offer_id: offer1.id)
        logs = Hermes::Log.only(:code, :action, :score_id).where(code: 'SC_C')
        expect(Hermes::HousingOffer.count).to be 2
        expect(logs.count).to be 8
        expect(Hermes::Score.unscoped.count).to be 8
        score_ids = scores.pluck(:id)
        expect(logs.pluck(:score_id)).to include(score_ids[0]).or(
          include(score_ids[1]).or(
            include(score_ids[2]).or(include(score_ids[3]))
          )
        )
        
        expect(scores.count).to eq(4)
        applicant1_score = scores.find_by(applicant: applicant1)
        applicant2_score = scores.find_by(applicant: applicant2)
        applicant3_score = scores.find_by(applicant: applicant3)
        applicant4_score = scores.find_by(applicant: applicant4)
        
        expect(applicant1_score.profile_epci_score).to be 30
        expect(applicant1_score.eligible_position).to be 3
        expect(applicant1_score.applicated_position).to be 1
        expect(applicant2_score.profile_epci_score).to be 40
        expect(applicant2_score.eligible_position).to be 2
        expect(applicant2_score.applicated_position).to be 1
        expect(applicant3_score.profile_epci_score).to be 30
        expect(applicant3_score.eligible_position).to be 4
        expect(applicant3_score.applicated_position).to be 1
        expect(applicant4_score.profile_epci_score).to be 40
        expect(applicant4_score.eligible_position).to be 1
        expect(applicant4_score.applicated_position).to be 1
        
        expect(offer1.reload.eligible_requests_nb).to be 4
        expect(offer1.reload.floor_category_1).to be 3050.0
      end
    end
  end
  
  describe 'Update realized_by_typologies of linked lessor_objective' do
    let!(:linked_lessor) { create(:lessor) }
    
    let!(:basic_housing_offer) do
      create(
        :housing_offer,
        lessor: linked_lessor,
        lessor_status: 'not rented',
        typology: 'T2',
        lessor_status_updated_at: DateTime.current # à supprimer une fois les statuts mis à jour 
      )
    end

    let!(:linked_lessor_objective) do
      create(
        :lessor_objective,
        activation_date: DateTime.current,
        lessor: linked_lessor,
        target_by_typologies: {:all =>100},
      )
    end

    it 'setting up' do
      linked_lessor_objective.reload
      expect(linked_lessor_objective.target_by_typologies).to(eq({'all'=>100}))
      expect(linked_lessor_objective.present_typologies).to(eq(['all']))
      expect(linked_lessor_objective.realized_by_typologies).to(eq({'all'=>0}))
    end

   
  end
  
  describe 'Realized calculation work at objective creation' do
    let!(:linked_lessor) { create(:lessor) }

    let!(:basic_housing_offer1) do
      create(
        :housing_offer,
        lessor: linked_lessor,
        typology: 'T2',
      )
    end

    let!(:status1) do
      create(
        :status,
        object_id: basic_housing_offer1.id.to_s,
        status_label: 'lessor_status',
        title: 'PLACE'
      )
    end

    let!(:basic_housing_offer2) do
      create(
        :housing_offer,
        lessor: linked_lessor,
        typology: 'T2',
      )
    end

    let!(:status2) do
      create(
        :status,
        object_id: basic_housing_offer2.id.to_s,
        status_label: 'lessor_status',
        title: 'PLACE', 
      )
    end

    let!(:linked_lessor_objective) do
      create(
        :lessor_objective,
        activation_date: DateTime.current,
        lessor: linked_lessor,
        target_by_typologies: {:T2 =>100},
      )
    end

    it 'count when offer placed' do
      expect(linked_lessor_objective.target_by_typologies).to(eq({:T2=>100}))
      expect(linked_lessor_objective.present_typologies).to(eq([:T2]))
      expect(linked_lessor_objective.realized_by_typologies).to(eq({:T2=>2}))
    end
  end

  describe 'Realized calculation work at objective creation within old env' do
    before :each do
      config = Setting.find_or_create_by(key: 'config', env: Rails.env)
      config.set(
        'values.transition_handler'.to_sym => 'old',
        'values.offer_family_slug'.to_sym => 'dev_housing_offers'
      )
      expect(config.reload.values.keys).to include('transition_handler').and(include('offer_family_slug'))
      expect(config.reload.values['transition_handler']).to eq('old')
      expect(config.reload.values['offer_family_slug']).to eq('dev_housing_offers')
    end

    let!(:linked_lessor) { create(:lessor) }

    let!(:basic_housing_offer1) do
      config = Setting.find_or_create_by(key: 'config', env: Rails.env)
      config.set(
        'values.transition_handler'.to_sym => 'old',
        'values.offer_family_slug'.to_sym => 'dev_housing_offers'
      )
      expect(config.reload.values.keys).to include('transition_handler').and(include('offer_family_slug'))
      expect(config.reload.values['transition_handler']).to eq('old')
      expect(config.reload.values['offer_family_slug']).to eq('dev_housing_offers')
      create(
        :housing_offer,
        lessor: linked_lessor,
        lessor_status: 'PLACE',
        typology: 'T2',
        lessor_status_updated_at: DateTime.current # TODO: à supprimer une fois les statuts mis à jour 
      )
    end

    let!(:basic_housing_offer2) do
      config = Setting.find_or_create_by(key: 'config', env: Rails.env)
      config.set(
        'values.transition_handler'.to_sym => 'old',
        'values.offer_family_slug'.to_sym => 'dev_housing_offers'
      )
      expect(config.reload.values.keys).to include('transition_handler').and(include('offer_family_slug'))
      expect(config.reload.values['transition_handler']).to eq('old')
      expect(config.reload.values['offer_family_slug']).to eq('dev_housing_offers')
      create(
        :housing_offer,
        lessor: linked_lessor,
        lessor_status: 'PLACE',
        typology: 'T2',
        lessor_status_updated_at: DateTime.current # TODO: à supprimer une fois les statuts mis à jour 
      )
    end
    let!(:basic_housing_offer3) do
      create(
          :housing_offer,
          lessor: linked_lessor,
          typology: 'T2',
          is_mad: true
      )
    end
    let!(:linked_lessor_objective) do
      config = Setting.find_or_create_by(key: 'config', env: Rails.env)
      config.set(
        'values.transition_handler'.to_sym => 'old',
        'values.offer_family_slug'.to_sym => 'dev_housing_offers'
      )
      expect(config.reload.values.keys).to include('transition_handler').and(include('offer_family_slug'))
      expect(config.reload.values['transition_handler']).to eq('old')
      expect(config.reload.values['offer_family_slug']).to eq('dev_housing_offers')
      create(
        :lessor_objective,
        activation_date: DateTime.current,
        lessor: linked_lessor,
        target_by_typologies: {:T2 =>100},
      )
    end

    it 'count when offer placed' do
      expect(linked_lessor_objective.target_by_typologies).to(eq({:T2=>100}))
      expect(linked_lessor_objective.present_typologies).to(eq([:T2]))
      expect(linked_lessor_objective.realized_by_typologies).to(eq({:T2=>2}))
    end

    it 'edit offer with nud trigger sopra flux' do
      after_save_res = basic_housing_offer3.after_save_actions
      #expect(after_save_res[:exectuted]).to eq(true)
    end
  end

  context 'with large quantity' do
    nb = 150
    nb.times do |i|
      let!("offer#{i}") { create(:housing_offer) }
    end

    describe 'als_id is set' do
      let(:all_offers) do
        Hermes::HousingOffer.only(:als_id)
                            .where(als_id: { :$ne => nil })
      end

      it "has #{nb} entities" do
        expect(all_offers.count).to be nb
      end

      let(:als_ids) { all_offers.pluck(:als_id).uniq.compact }

      it 'produces valid lenght' do
        expect(als_ids.map(&:length)).to all(be(18))
      end

      it 'has no duplicates' do
        expect(
          als_ids.map { |als_id| all_offers.where(als_id: als_id).count }
        ).to all(be(1))
      end
    end
  end

  describe 'on creation PUBLIE' do
    let!(:offer_1) do
      create(
        :housing_offer,
        typology: 'T2',
        rent_amount: 434.32,
        rental_charges: 53.12,
        parking_charges: 22.43,
        surface: 34,
        offer_status: 'PUBLIE',
        threshold_type: 'PLA LM',
        department: '34',
        threshold_subtype: 'Autres régions'
      )
    end

    it 'eligible_query ' do
      expect(offer_1.reload.eligibility_query).not_to be_nil;
    end

  end
end
